os-common.o os-common.d : os-common.c genesis/sbcl.h globals.h runtime.h lispobj.h \
 align.h os.h target-os.h target-arch-os.h target-arch.h genesis/cons.h \
 lispobj.h genesis/hash-table.h genesis/vector.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/static-symbols.h thread.h genesis/thread.h gencgc-alloc-region.h \
 interrupt.h validate.h gc-typedefs.h arch.h interr.h immobile-space.h \
 code.h genesis/closure.h gc-assert.h search.h sys_mmap.inc
