arm-linux-os.o arm-linux-os.d : arm-linux-os.c genesis/sbcl.h os.h runtime.h lispobj.h \
 align.h target-os.h target-arch-os.h target-arch.h arch.h thread.h \
 globals.h genesis/symbol.h lispobj.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h genesis/vector.h interrupt.h \
 validate.h gc-typedefs.h interr.h lispregs.h target-lispregs.h \
 genesis/regnames.h
