/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_REGNAMES
#define SBCL_GENESIS_REGNAMES
#define REGNAMES \
    "R0", "R1", "R2", "LEXENV", "NL2", "CSP", "NL3", "OCFP", "R8", "NFP",\
    "NULL", "CFP", "NARGS", "NSP", "LR", "PC"
#define BOXED_REGISTERS { \
    reg_R0, reg_R1, reg_R2, reg_LEXENV, reg_R8\
}
#endif
