/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_INSTANCE
#define SBCL_GENESIS_INSTANCE
#ifdef __ASSEMBLER__

/* These offsets are SLOT-OFFSET * N-WORD-BYTES - LOWTAG
 * so they work directly on tagged addresses. */

#define INSTANCE_SLOTS_OFFSET 3
#define INSTANCE_SIZE 1
#define FUNCALLABLE_INSTANCE_TRAMPOLINE_OFFSET -1
#define FUNCALLABLE_INSTANCE_FUNCTION_OFFSET 3
#define FUNCALLABLE_INSTANCE_LAYOUT_OFFSET 7
#define FUNCALLABLE_INSTANCE_INFO_OFFSET 11
#define FUNCALLABLE_INSTANCE_SIZE 4

#else /* __ASSEMBLER__ */

#include "lispobj.h"
struct instance {
    lispobj header;
    lispobj slots[1];
};
static inline struct instance* INSTANCE(lispobj obj) {
  return (struct instance*)(obj - 1);
}
struct funcallable_instance {
    lispobj header;
    lispobj trampoline;
    lispobj function;
    lispobj layout;
    lispobj info[1];
};

#define LAYOUT_LAYOUT_ID 3
#define LFLIST_NODE_LAYOUT_ID 4
#define BROTHERTREE_UNARY_NODE_LAYOUT_ID 5
#define FD_STREAM_BUFFER_LAYOUT_ID 2
#define HASHSET_LAYOUT_ID 9
#define HASHSET_STORAGE_LAYOUT_ID 10
#define FINALIZER_NODE_LAYOUT_ID 13
#define PACKAGE_LAYOUT_ID 8
#define HASH_TABLE_LAYOUT_ID 7

struct layout {
    lispobj header; // = word_0_
    lispobj _layout;
    lispobj uw_flags;
    lispobj clos_hash;
    lispobj vtable;
    lispobj classoid;
    lispobj invalid;
    lispobj inherits;
    lispobj depthoid;
    lispobj length;
    lispobj _info;
    lispobj equalp_impl;
    lispobj slot_mapper;
    lispobj slot_table;
    lispobj uw_id_word0;
    lispobj uw_id_word1;
    lispobj uw_id_word2;
    lispobj uw_id_word3;
    lispobj uw_id_word4;
    lispobj uw_id_word5;
};
static inline struct layout* LAYOUT(lispobj obj) {
  return (struct layout*)(obj - 1);
}
#include "instance.inc"

#endif /* __ASSEMBLER__ */
#endif
